import imaplib
import logging
import poplib

from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from ..classes import SourceBackend
from ..exceptions import SourceException

from .literals import (
    DEFAULT_EMAIL_IMAP_MAILBOX, DEFAULT_EMAIL_IMAP_SEARCH_CRITERIA,
    DEFAULT_EMAIL_IMAP_STORE_COMMANDS, DEFAULT_EMAIL_POP3_TIMEOUT,
    SOURCE_INTERVAL_UNCOMPRESS_CHOICES
)
from .mixins import (
    SourceBackendCompressedMixin, SourceBackendEmailMixin,
    SourceBackendPeriodicMixin, SourceBaseMixin
)

__all__ = ('SourceBackendIMAPEmail', 'SourceBackendPOP3Email')
logger = logging.getLogger(name=__name__)


class SourceBackendIMAPEmail(
    SourceBackendCompressedMixin, SourceBackendEmailMixin,
    SourceBackendPeriodicMixin, SourceBaseMixin, SourceBackend
):
    field_order = (
        'mailbox', 'search_criteria', 'store_commands', 'execute_expunge',
        'mailbox_destination'
    )
    fields = {
        'mailbox': {
            'class': 'django.forms.fields.CharField',
            'default': DEFAULT_EMAIL_IMAP_MAILBOX,
            'help_text': _('IMAP Mailbox from which to check for messages.'),
            'kwargs': {
                'max_length': 64,
            },
            'label': _('Mailbox')
        },
        'search_criteria': {
            'blank': True,
            'class': 'django.forms.fields.CharField',
            'default': DEFAULT_EMAIL_IMAP_SEARCH_CRITERIA,
            'help_text': _(
                'Criteria to use when searching for messages to process. '
                'Use the format specified in '
                'https://tools.ietf.org/html/rfc2060.html#section-6.4.4'
            ),
            'label': _('Search criteria'),
            'null': True,
        },
        'store_commands': {
            'blank': True,
            'class': 'django.forms.fields.CharField',
            'default': DEFAULT_EMAIL_IMAP_STORE_COMMANDS,
            'help_text': _(
                'IMAP STORE command to execute on messages after they are '
                'processed. One command per line. Use the commands specified in '
                'https://tools.ietf.org/html/rfc2060.html#section-6.4.6 or '
                'the custom commands for your IMAP server.'
            ),
            'label': _('Store commands'),
            'null': True,
        },
        'execute_expunge': {
            'class': 'django.forms.fields.BooleanField',
            'default': True,
            'help_text': _(
                'Execute the IMAP expunge command after processing each email '
                'message.'
            ),
            'label': _('Execute expunge'),
            'required': False
        },
        'mailbox_destination': {
            'blank': True,
            'class': 'django.forms.fields.CharField',
            'help_text': _(
                'IMAP Mailbox to which processed messages will be copied.'
            ),
            'label': _('Destination mailbox'),
            'max_length': 96,
            'null': True,
            'required': False
        }
    }
    label = _('IMAP email')
    uncompress_choices = SOURCE_INTERVAL_UNCOMPRESS_CHOICES
    widgets = {
        'search_criteria': {
            'class': 'django.forms.widgets.Textarea',
        },
        'store_commands': {
            'class': 'django.forms.widgets.Textarea',
        }
    }

    def get_shared_uploaded_files(self):
        dry_run = self.process_kwargs.get('dry_run', False)

        logger.debug(msg='Starting IMAP email fetch')
        logger.debug('host: %s', self.kwargs['host'])
        logger.debug('ssl: %s', self.kwargs['ssl'])

        if self.kwargs['ssl']:
            server = imaplib.IMAP4_SSL(
                host=self.kwargs['host'], port=self.kwargs['port']
            )
        else:
            server = imaplib.IMAP4(
                host=self.kwargs['host'], port=self.kwargs['port']
            )

        server.login(
            user=self.kwargs['username'], password=self.kwargs['password']
        )
        try:
            server.select(mailbox=self.kwargs['mailbox'])
        except Exception as exception:
            raise SourceException(
                'Error selecting mailbox: {}; {}'.format(
                    self.kwargs['mailbox'], exception
                )
            )

        try:
            status, data = server.uid(
                'SEARCH', None, *self.kwargs['search_criteria'].strip().split()
            )
        except Exception as exception:
            raise SourceException(
                'Error executing search command; {}'.format(exception)
            )

        if data:
            # data is a space separated sequence of message uids.
            uids = data[0].split()
            logger.debug('messages count: %s', len(uids))
            logger.debug('message uids: %s', uids)

            for uid in uids:
                logger.debug('message uid: %s', uid)

                try:
                    status, data = server.uid('FETCH', uid, '(RFC822)')
                except Exception as exception:
                    raise SourceException(
                        'Error fetching message uid: {}; {}'.format(
                            uid, exception
                        )
                    )

                try:
                    shared_uploaded_files = self.process_message(
                        message=data[0][1]
                    )
                except Exception as exception:
                    raise SourceException(
                        'Error processing message uid: {}; {}'.format(
                            uid, exception
                        )
                    )

                if not dry_run:
                    if self.kwargs['store_commands']:
                        for command in self.kwargs['store_commands'].split('\n'):
                            try:
                                args = [uid]
                                args.extend(command.strip().split(' '))
                                server.uid('STORE', *args)
                            except Exception as exception:
                                raise SourceException(
                                    'Error executing IMAP store command "{}" '
                                    'on message uid {}; {}'.format(
                                        command, uid, exception
                                    )
                                )

                    if self.kwargs['mailbox_destination']:
                        try:
                            server.uid(
                                'COPY', uid, self.kwargs['mailbox_destination']
                            )
                        except Exception as exception:
                            raise SourceException(
                                'Error copying message uid {} to mailbox {}; '
                                '{}'.format(
                                    uid, self.kwargs['mailbox_destination'], exception
                                )
                            )

                    if self.kwargs['execute_expunge']:
                        server.expunge()

        server.close()
        server.logout()

        return shared_uploaded_files


class SourceBackendPOP3Email(
    SourceBackendCompressedMixin, SourceBackendEmailMixin,
    SourceBackendPeriodicMixin, SourceBaseMixin, SourceBackend
):
    field_order = ('timeout', 'uncompress',)
    fields = {
        'timeout': {
            'class': 'django.forms.fields.IntegerField',
            'default': DEFAULT_EMAIL_POP3_TIMEOUT,
            'kwargs': {
                'min_value': 0
            },
            'label': _('Timeout')
        }
    }
    label = _('POP3 email')

    def get_shared_uploaded_files(self):
        dry_run = self.process_kwargs.get('dry_run', False)

        logger.debug(msg='Starting POP3 email fetch')
        logger.debug('host: %s', self.kwargs['host'])
        logger.debug('ssl: %s', self.kwargs['ssl'])

        if self.kwargs['ssl']:
            server = poplib.POP3_SSL(host=self.kwargs['host'], port=self.kwargs['port'])
        else:
            server = poplib.POP3(
                host=self.kwargs['host'], port=self.kwargs['port'], timeout=self.kwargs['timeout']
            )

        server.getwelcome()
        server.user(self.kwargs['username'])
        server.pass_(self.kwargs['password'])

        messages_info = server.list()

        logger.debug(msg='messages_info:')
        logger.debug(msg=messages_info)
        logger.debug('messages count: %s', len(messages_info[1]))

        for message_info in messages_info[1]:
            message_number, message_size = message_info.split()
            message_number = int(message_number)

            logger.debug('message_number: %s', message_number)
            logger.debug('message_size: %s', message_size)

            message_lines = server.retr(which=message_number)[1]
            message_complete = force_text(s=b'\n'.join(message_lines))

            shared_uploaded_files = self.process_message(
                message=message_complete
            )
            if not dry_run:
                server.dele(which=message_number)

        server.quit()

        return shared_uploaded_files



import re

from django.db import migrations


def operation_update_event_types_names(apps, schema_editor):
    Action = apps.get_model(app_label='actstream', model_name='Action')
    StoredEventType = apps.get_model(
        app_label='events', model_name='StoredEventType'
    )

    known_namespaces = {
        'documents_': 'documents.',
        'checkouts_': 'checkouts.',
        'document_comment_': 'document_comments.',
        'parsing_document_': 'document_parsing.',
        'ocr_': 'ocr.',
        'tag_': 'tags.',
    }

    pattern = re.compile('|'.join(known_namespaces.keys()))

    for event_type in StoredEventType.objects.using(alias=schema_editor.connection.alias).all():
        event_type.name = pattern.sub(
            lambda x: known_namespaces[x.group()], event_type.name
        )
        event_type.save()

    for action in Action.objects.using(alias=schema_editor.connection.alias).all():
        action.verb = pattern.sub(
            lambda x: known_namespaces[x.group()], action.verb
        )
        action.save()


def operation_revert_event_types_names(apps, schema_editor):
    Action = apps.get_model(app_label='actstream', model_name='Action')
    StoredEventType = apps.get_model(
        app_label='events', model_name='StoredEventType'
    )

    known_namespaces = {
        r'documents\.': 'documents_',
        r'checkouts\.': 'checkouts_',
        r'document_comments\.': 'document_comment_',
        r'document_parsing\.': 'parsing_document_',
        r'ocr\.': 'ocr_',
        r'tags\.': 'tag_',
    }

    pattern = re.compile('|'.join(known_namespaces.keys()))

    for event_type in StoredEventType.objects.using(alias=schema_editor.connection.alias).all():
        old_name = event_type.name
        new_name = pattern.sub(
            lambda x: known_namespaces[x.group().replace('.', '\\.')],
            event_type.name
        )
        event_type.name = new_name
        if old_name == new_name:
            event_type.delete()
        else:
            event_type.save()

    for action in Action.objects.using(alias=schema_editor.connection.alias).all():
        new_name = pattern.sub(
            lambda x: known_namespaces[x.group().replace('.', '\\.')],
            action.verb
        )
        action.verb = new_name
        action.save()


class Migration(migrations.Migration):
    dependencies = [
        ('events', '0004_auto_20170731_0423'),
        ('actstream', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(
            code=operation_update_event_types_names,
            reverse_code=operation_revert_event_types_names
        ),
    ]



import logging

from django.db import models

logger = logging.getLogger(name=__name__)


class RoleManager(models.Manager):
    def get_by_natural_key(self, label):
        return self.get(label=label)


class StoredPermissionManager(models.Manager):
    def get_by_natural_key(self, namespace, name):
        return self.get(namespace=namespace, name=name)

    def purge_obsolete(self):
        total = 0

        for permission in self.all():
            try:
                permission.volatile_permission
            except KeyError:
                permission.delete()
                total += 1

        return total



import logging

from django.apps import apps
from django.contrib.auth import get_user_model

from mayan.apps.lock_manager.exceptions import LockError
from mayan.celery import app

logger = logging.getLogger(name=__name__)


@app.task(bind=True, ignore_result=True)
def task_cache_partition_purge(
    self, cache_partition_id, content_type_id=None, object_id=None,
    user_id=None
):
    CachePartition = apps.get_model(
        app_label='file_caching', model_name='CachePartition'
    )
    ContentType = apps.get_model(
        app_label='contenttypes', model_name='ContentType'
    )
    User = get_user_model()

    cache_partition = CachePartition.objects.get(pk=cache_partition_id)

    if content_type_id and object_id:
        content_type = ContentType.objects.get(pk=content_type_id)
        content_object = content_type.get_object_for_this_type(pk=object_id)
    else:
        content_object = None

    if user_id:
        user = User.objects.get(pk=user_id)
    else:
        user = None

    logger.info('Starting cache partition id %s purge', cache_partition)

    try:
        cache_partition._event_action_object = content_object
        cache_partition._event_actor = user
        cache_partition.purge()
    except LockError as exception:
        raise self.retry(exc=exception)
    else:
        logger.info('Finished cache partition id %s purge', cache_partition)


@app.task(bind=True, ignore_result=True)
def task_cache_purge(self, cache_id, user_id=None):
    Cache = apps.get_model(
        app_label='file_caching', model_name='Cache'
    )
    User = get_user_model()

    cache = Cache.objects.get(pk=cache_id)
    if user_id:
        user = User.objects.get(pk=user_id)
    else:
        user = None

    logger.info('Starting cache id %s purge', cache)
    try:
        cache._event_actor = user
        cache._event_keep_attributes = ('_event_actor',)
        cache.purge()
    except LockError as exception:
        raise self.retry(exc=exception)
    else:
        logger.info('Finished cache id %s purge', cache)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.documents.models.document_version_models import DocumentVersion

DEFAULT_DOCUMENT_BODY_TEMPLATE = _(
    'Attached to this email is the {{ object_name }}: {{ object }}\n\n '
    '--------\n '
    'This email has been sent from %(project_title)s (%(project_website)s)'
)
DEFAULT_DOCUMENT_SUBJECT_TEMPLATE = _('{{ object_name }}: {{ object }}')
DEFAULT_LINK_BODY_TEMPLATE = _(
    'To access this {{ object_name }} click on the following link: '
    '{{ link }}\n\n--------\n '
    'This email has been sent from %(project_title)s (%(project_website)s)'
)
DEFAULT_LINK_SUBJECT_TEMPLATE = _('Link for {{ object_name }}: {{ object }}')
EMAIL_SEPARATORS = (',', ';')

DOCUMENT_FILE_CONTENT_FUNCTION_DOTTED_PATH = 'mayan.apps.mailer.utils.get_document_file_content'
DOCUMENT_FILE_MIME_TYPE_FUNCTION_DOTTED_PATH = 'mayan.apps.mailer.utils.get_document_file_mime_type'

DOCUMENT_VERSION_CONTENT_FUNCTION_DOTTED_PATH = 'mayan.apps.mailer.utils.get_document_version_content'
DOCUMENT_VERSION_MIME_TYPE_FUNCTION_DOTTED_PATH = 'mayan.apps.mailer.utils.get_document_version_mime_type'

MODEL_SEND_FUNCTION_DOTTED_PATH = {
    DocumentFile: {
        'content_function_dotted_path': DOCUMENT_FILE_CONTENT_FUNCTION_DOTTED_PATH,
        'mime_type_function_dotted_path': DOCUMENT_FILE_MIME_TYPE_FUNCTION_DOTTED_PATH
    },
    DocumentVersion: {
        'content_function_dotted_path': DOCUMENT_VERSION_CONTENT_FUNCTION_DOTTED_PATH,
        'mime_type_function_dotted_path': DOCUMENT_VERSION_MIME_TYPE_FUNCTION_DOTTED_PATH
    }
}



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('document_parsing', '0009_auto_20201012_0708'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='documentfilepagecontent',
            options={'verbose_name': 'Document file page content', 'verbose_name_plural': 'Document file page contents'},
        ),
        migrations.AlterField(
            model_name='documentfilepagecontent',
            name='document_file_page',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='content', to='documents.DocumentFilePage', verbose_name='Document file page'),
        ),
    ]



import os

from django.conf import settings

TEST_DOWNLOAD_FILE_CONTENT_FILE_NAME = 'test content name'

TEST_CONTENT = 'testcontent'
TEST_FILE_NAME = 'test_file'

# Filenames
TEST_ARCHIVE_MSG_STRANGE_DATE_FILENAME = 'strangeDate.msg'
TEST_ARCHIVE_ZIP_SPECIAL_CHARACTERS_FILENAME_MEMBER = 'test_archvive_with_special_characters_filename_member.zip'
TEST_ARCHIVE_ZIP_CP437_MEMBER = 'test_archvive_with_cp437_member.zip'
TEST_FILENAME1 = 'test_file1.txt'
TEST_FILENAME2 = 'test_file2.txt'
TEST_FILENAME3 = 'test_file3.txt'
TEST_FILE_CONTENTS_1 = b'TEST FILE 1\n'
TEST_FILE_CONTENTS_2 = 'TEST FILE 2\n'
TEST_TAR_BZ2_FILE = 'test_file.tar.bz2'
TEST_TAR_FILE = 'test_file.tar'
TEST_TAR_GZ_FILE = 'test_file.tar.gz'
TEST_ZIP_FILE = 'test_file.zip'
TEST_COMPRESSED_FILE_CONTENTS = [TEST_FILENAME1, TEST_FILENAME2]

# File paths
TEST_ARCHIVE_ZIP_SPECIAL_CHARACTERS_FILENAME_MEMBER_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'common', 'tests', 'contrib',
    TEST_ARCHIVE_ZIP_SPECIAL_CHARACTERS_FILENAME_MEMBER
)
TEST_ARCHIVE_ZIP_CP437_MEMBER_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'common', 'tests', 'contrib',
    TEST_ARCHIVE_ZIP_CP437_MEMBER
)
TEST_FILE3_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'common', 'tests', 'contrib', TEST_FILENAME3
)
TEST_TAR_BZ2_FILE_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'common', 'tests', 'contrib', TEST_TAR_BZ2_FILE
)
TEST_TAR_FILE_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'common', 'tests', 'contrib', TEST_TAR_FILE
)
TEST_TAR_GZ_FILE_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'common', 'tests', 'contrib', TEST_TAR_GZ_FILE
)
TEST_ZIP_FILE_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'common', 'tests', 'contrib', TEST_ZIP_FILE
)
TEST_ARCHIVE_MSG_STRANGE_DATE_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'common', 'tests', 'contrib',
    TEST_ARCHIVE_MSG_STRANGE_DATE_FILENAME
)

TEST_SHARED_UPLOADED_FILE_FILENAME = 'test_shared_uploaded_file_filename'



from django.utils.encoding import force_bytes

TEST_EMAIL_ATTACHMENT_AND_INLINE = '''Subject: Test 03: inline and attachments
To: Renat Gilmanov
Content-Type: multipart/mixed; boundary=001a11c24d809f1525051712cc78

--001a11c24d809f1525051712cc78
Content-Type: multipart/related; boundary=001a11c24d809f1523051712cc77

--001a11c24d809f1523051712cc77
Content-Type: text/html; charset=UTF-8
Content-Transfer-Encoding: quoted-printable

<div dir=3D"ltr">Lorem ipsum dolor sit amet, consectetur adipiscing elit. P=
ellentesque odio urna, bibendum eu ultricies in, dignissim in magna. Vivamu=
s risus justo, viverra sed dapibus eu, laoreet eget erat. Sed pretium a urn=
a id pulvinar.<br><br><img src=3D"cid:ii_ia6yyemg0_14d9636d8ac7a587" height=
=3D"218" width=3D"320"><br>=E2=80=8B<br>Cras eu velit ac purus feugiat impe=
rdiet nec sit amet ipsum. Praesent gravida lobortis justo, nec tristique ve=
lit sagittis finibus. Suspendisse porta ante id diam varius, in cursus ante=
 luctus. Aenean a mollis mi. Pellentesque accumsan lacus sed erat vulputate=
, et semper tellus condimentum.<br><br>Best regards</div>

--001a11c24d809f1523051712cc77
Content-Type: image/png; name="test-01.png"
Content-Disposition: inline; filename="test-01.png"
Content-Transfer-Encoding: base64
Content-ID: <ii_ia6yyemg0_14d9636d8ac7a587>
X-Attachment-Id: ii_ia6yyemg0_14d9636d8ac7a587

iVBORw0KGgoAAAANSUhEUgAAAUAAAADaCAYAAADXGps7AAAABHNCSVQICAgIfAhkiAAAAAlwSFlz
AAALewAAC3sBSRnwgAAAABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAALnSURB
...
QCDLAIEsAwSyDBDIMkAgywCBLAMEsgwQyDJAIMsAgSwDBLIMEMgyQCDLAIEsAwSyDBDIMkAg6wK+
4gU280YtuwAAAABJRU5ErkJggg==
--001a11c24d809f1523051712cc77--
--001a11c24d809f1525051712cc78
Content-Type: image/png; name="test-02.png"
Content-Disposition: attachment; filename="test-02.png"
Content-Transfer-Encoding: base64
X-Attachment-Id: f_ia6yymei1'''
TEST_EMAIL_BASE64_FILENAME_ATTACHMENT_FILENAME = 'Ampelmännchen.txt'
TEST_EMAIL_BASE64_MESSAGE_ID = '<00000001.465619c9.1.00@BRN30055CCF4D76>'
TEST_EMAIL_BASE64_FILENAME = force_bytes(s='''From: noreply@example.com
To: test@example.com
Subject: Scan to E-mail Server Job
Date: Tue, 23 May 2017 23:03:37 +0200
Message-Id: {}
Mime-Version: 1.0
Content-Type: multipart/mixed;
    boundary="RS1tYWlsIENsaWVudA=="
X-Mailer: E-mail Client

This is multipart message.

--RS1tYWlsIENsaWVudA==
Content-Type: text/plain; charset=iso-8859-1
Content-Transfer-Encoding: quoted-printable

Sending device cannot receive e-mail replies.
--RS1tYWlsIENsaWVudA==
Content-Type: text/plain
Content-Transfer-Encoding: base64
Content-Disposition: attachment; filename="=?UTF-8?B?QW1wZWxtw6RubmNoZW4udHh0?="

SGFsbG8gQW1wZWxtw6RubmNoZW4hCg==

--RS1tYWlsIENsaWVudA==--'''.format(TEST_EMAIL_BASE64_MESSAGE_ID))
TEST_EMAIL_BASE64_FILENAME_FROM = 'noreply@example.com'
TEST_EMAIL_BASE64_FILENAME_SUBJECT = 'Scan to E-mail Server Job'
TEST_EMAIL_NO_CONTENT_TYPE = '''MIME-Version: 1.0
Received: by 10.0.0.1 with HTTP; Mon, 9 Apr 2018 00:00:00 -0400 (AST)
X-Originating-IP: [10.0.0.1]
Date: Mon, 9 Apr 2018 0:00:0 -0400
Delivered-To: test-sender@example.com
Message-ID: <CAEAsyCbSF1Bk7CBuu6zp3Qs8=j2iUkNi3dPkGe6z40q4dmaogQ@mail.gmail.com>
Subject: Test message with no content type
From: Test Sender <test-sender@example.com>
To: test-receiver@example.com

Test email without a content type'''
TEST_EMAIL_NO_CONTENT_TYPE_STRING = b'Test email without a content type'
TEST_EMAIL_INLINE_IMAGE = '''Subject: Test 01: inline only
To: Renat Gilmanov
Content-Type: multipart/related; boundary=089e0149bb0ea4e55c051712afb5

--089e0149bb0ea4e55c051712afb5
Content-Type: text/html; charset=UTF-8
Content-Transfer-Encoding: quoted-printable

<div dir=3D"ltr">Lorem ipsum dolor sit amet, consectetur adipiscing elit. P=
ellentesque odio urna, bibendum eu ultricies in, dignissim in magna. Vivamu=
s risus justo, viverra sed dapibus eu, laoreet eget erat. Sed pretium a urn=
a id pulvinar.<br><br><img src=3D"cid:ii_ia6yo3z92_14d962f8450cc6f1" height=
=3D"218" width=3D"320"><br>=E2=80=8B<br>Cras eu velit ac purus feugiat impe=
rdiet nec sit amet ipsum. Praesent gravida lobortis justo, nec tristique ve=
lit sagittis finibus. Suspendisse porta ante id diam varius, in cursus ante=
 luctus. Aenean a mollis mi. Pellentesque accumsan lacus sed erat vulputate=
, et semper tellus condimentum.<br><br>Best regards<br></div>

--089e0149bb0ea4e55c051712afb5
Content-Type: image/png; name="test-01.png"
Content-Disposition: inline; filename="test-01.png"
Content-Transfer-Encoding: base64
Content-ID: <ii_ia6yo3z92_14d962f8450cc6f1>
X-Attachment-Id: ii_ia6yo3z92_14d962f8450cc6f1

iVBORw0KGgoAAAANSUhEUgAAAUAAAADaCAYAAADXGps7AAAABHNCSVQICAgIfAhkiAAAAAlwSFlz
AAALewAAC3sBSRnwgAAAABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAALnSURB
...
QCDLAIEsAwSyDBDIMkAgywCBLAMEsgwQyDJAIMsAgSwDBLIMEMgyQCDLAIEsAwSyDBDIMkAg6wK+
4gU280YtuwAAAABJRU5ErkJggg==
--089e0149bb0ea4e55c051712afb5--'''
TEST_EMAIL_ZERO_LENGTH_ATTACHMENT = '''From rosarior@localhost Tue Apr  2 23:40:26 2019
Received: from rosarior (uid 1000)
    (envelope-from rosarior@localhost)
    id 2011e6
    by localhost (DragonFly Mail Agent v0.11);
    Tue, 02 Apr 2019 23:40:26 -0400
Message-ID: <22514.1554262827@localhost>
Mime-Version: 1.0
To: rosarior@localhost
Subject: mpack
Content-Type: multipart/mixed; boundary="-"
Date: Tue, 02 Apr 2019 23:40:26 -0400
From: <rosarior@localhost>

This is a MIME encoded message.  Decode it with "munpack"
or any other MIME reading software.  Mpack/munpack is available
via anonymous FTP in ftp.andrew.cmu.edu:pub/mpack/
---
Content-Type: application/octet-stream; name="zero"
Content-Transfer-Encoding: base64
Content-Disposition: inline; filename="zero"
Content-MD5: 1B2M2Y8AsgTpgAmY7PhCfg==


-----'''
TEST_SOURCE_BACKEND_EMAIL_PATH = 'mayan.apps.sources.tests.source_backends.SourceBackendTestEmail'
TEST_SOURCE_BACKEND_PERIODIC_PATH = 'mayan.apps.sources.tests.source_backends.SourceBackendTestPeriodic'
TEST_SOURCE_BACKEND_PATH = 'mayan.apps.sources.tests.source_backends.SourceBackendSimple'
TEST_SOURCE_LABEL = 'test source'
TEST_SOURCE_LABEL_EDITED = 'test source edited'

TEST_SOURCES_STAGING_FILE_CACHE_STORAGE_BACKEND = 'test.backend'
TEST_SOURCES_STAGING_FILE_CACHE_STORAGE_BACKEND_ARGUMENTS = {'location': 'test value'}

TEST_STAGING_PREVIEW_HEIGHT = ''
TEST_STAGING_PREVIEW_WIDTH = 640

TEST_WATCHFOLDER_SUBFOLDER = 'test_subfolder'



import logging
from pathlib import Path

from django.contrib.auth import get_user_model
from django.core.files import File
from django.db import models, transaction
from django.utils.translation import ugettext_lazy as _

from mayan.apps.converter.layers import layer_saved_transformations
from mayan.apps.databases.model_mixins import (
    BackendModelMixin, ExtraDataModelMixin
)
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.tasks import task_document_upload
from mayan.apps.events.classes import EventManagerSave
from mayan.apps.events.decorators import method_event
from mayan.apps.storage.compressed_files import Archive
from mayan.apps.storage.exceptions import NoMIMETypeMatch
from mayan.apps.storage.models import SharedUploadedFile

from .classes import SourceBackendNull
from .events import event_source_created, event_source_edited
from .managers import SourceManager

logger = logging.getLogger(name=__name__)


class Source(BackendModelMixin, ExtraDataModelMixin, models.Model):
    _backend_model_null_backend = SourceBackendNull

    label = models.CharField(
        db_index=True, help_text=_('A short text to describe this source.'),
        max_length=128, unique=True, verbose_name=_('Label')
    )
    enabled = models.BooleanField(default=True, verbose_name=_('Enabled'))

    objects = SourceManager()

    class Meta:
        ordering = ('label',)
        verbose_name = _('Source')
        verbose_name_plural = _('Sources')

    @staticmethod
    def callback_post_task_document_upload(document_file, **kwargs):
        source = Source.objects.get(pk=kwargs['source_id'])

        if kwargs['user_id']:
            User = get_user_model()
            try:
                user = User.objects.get(pk=kwargs['user_id'])
            except User.DoesNotExist:
                """Skip adding recent document."""
            else:
                document_file.document.add_as_recent_document_for_user(
                    user=user
                )

        layer_saved_transformations.copy_transformations(
            source=source, targets=document_file.pages.all()
        )

        source.get_backend_instance().callback(
            document_file=document_file, **kwargs
        )

    def __str__(self):
        return '%s' % self.label

    def delete(self, *args, **kwargs):
        with transaction.atomic():
            self.get_backend_instance().delete()
            super().delete(*args, **kwargs)

    def execute_action(self, name, **kwargs):
        return self.get_backend_instance().execute_action(name=name, **kwargs)

    def fullname(self):
        return '{} {}'.format(self.get_backend_label(), self.label)

    def get_action(self, name):
        return self.get_backend().get_action(name=name)

    def get_actions(self):
        return self.get_backend().get_actions()

    def handle_file_object_upload(
        self, document_type, file_object, callback_kwargs=None,
        description=None, expand=False, label=None, language=None, user=None
    ):
        """
        Handle an upload request from a file object which may be an individual
        document or a compressed file containing multiple documents.
        """
        callback_kwargs = callback_kwargs or {}

        if expand:
            try:
                compressed_file = Archive.open(file_object=file_object)
                for compressed_file_member in compressed_file.members():
                    with compressed_file.open_member(filename=compressed_file_member) as compressed_file_member_file_object:
                        # Recursive call to expand nested compressed files
                        # expand=True literal for recursive nested files.
                        # Might cause problem with office files inside a
                        # compressed file.
                        self.handle_file_object_upload(
                            callback_kwargs=callback_kwargs,
                            document_type=document_type,
                            description=description,
                            expand=False,
                            file_object=compressed_file_member_file_object,
                            # Use the filename only and not the whole path.
                            label=Path(compressed_file_member).name,
                            language=language,
                            user=user
                        )

                # Avoid executing the expand=False code path.
                return
            except NoMIMETypeMatch:
                logger.debug(msg='No expanding; Exception: NoMIMETypeMatch')
                # Fallthrough to same code path as expand=False to avoid
                # duplicating code.

        shared_uploaded_file = SharedUploadedFile.objects.create(
            file=File(file_object)
        )

        Document.execute_pre_create_hooks(
            kwargs={
                'document_type': document_type,
                'user': user
            }
        )

        if user:
            user_id = user.pk
        else:
            user_id = None

        final_callback_kwargs = callback_kwargs.copy()
        final_callback_kwargs.update(
            {
                'source_id': self.pk,
                'user_id': user_id
            }
        )

        task_document_upload.apply_async(
            kwargs={
                'document_type_id': document_type.pk,
                'shared_uploaded_file_id': shared_uploaded_file.pk,
                'description': description,
                'label': label,
                'language': language,
                'user_id': user_id,
                'callback_dotted_path': 'mayan.apps.sources.models.Source',
                'callback_function': 'callback_post_task_document_upload',
                'callback_kwargs': final_callback_kwargs
            }
        )

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_source_created,
            'target': 'self',
        },
        edited={
            'event': event_source_edited,
            'target': 'self',
        }
    )
    def save(self, *args, **kwargs):
        is_new = not self.pk

        with transaction.atomic():
            super().save(*args, **kwargs)

            self.get_backend_instance().clean()

            if is_new:
                self.get_backend_instance().create()
            else:
                self.get_backend_instance().save()



from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..permissions import permission_resolved_smart_link_view

from .mixins import (
    ResolvedSmartLinkDocumentViewTestMixin, SmartLinkTestMixin
)


class ResolvedSmartLinkDocumentViewTestCase(
    SmartLinkTestMixin, ResolvedSmartLinkDocumentViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()

        self._create_test_document_stub()
        self._create_test_document_stub(label='linked')
        self._create_test_smart_link(add_test_document_type=True)
        self._create_test_smart_link_condition()

    def test_document_resolved_smart_link_list_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_resolved_smart_link_list_view()
        self.assertNotContains(
            response=response, status_code=404,
            text=self.test_documents[0].label
        )
        self.assertNotContains(
            response=response, status_code=404,
            text=self.test_documents[0].uuid
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_resolved_smart_link_list_view_with_document_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_test_document_resolved_smart_link_list_view()
        self.assertContains(
            response=response, status_code=200,
            text=self.test_documents[0].label
        )
        self.assertNotContains(
            response=response, status_code=200,
            text=self.test_documents[0].uuid
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_resolved_smart_link_list_view_with_smart_link_access(self):
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_test_document_resolved_smart_link_list_view()
        self.assertNotContains(
            response=response, status_code=404,
            text=self.test_documents[0].label
        )
        self.assertNotContains(
            response=response, status_code=404,
            text=self.test_documents[0].uuid
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_resolved_smart_link_list_view_with_full_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_test_document_resolved_smart_link_list_view()
        self.assertContains(
            response=response, status_code=200,
            text=self.test_documents[0].label
        )
        self.assertContains(
            response=response, status_code=200,
            text=self.test_documents[0].uuid
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_resolved_smart_link_list_view_with_full_access(self):
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )

        self.test_documents[0].delete()

        self._clear_events()

        response = self._request_test_document_resolved_smart_link_list_view()
        self.assertNotContains(
            response=response, status_code=404,
            text=self.test_documents[0].label
        )
        self.assertNotContains(
            response=response, status_code=404,
            text=self.test_documents[0].uuid
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_resolved_smart_document_list_with_no_permission(self):
        self._clear_events()

        response = self._request_test_document_resolved_smart_link_document_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_resolved_smart_document_list_with_main_document_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_test_document_resolved_smart_link_document_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_resolved_smart_document_list_with_smart_link_access(self):
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_test_document_resolved_smart_link_document_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_resolved_smart_document_list_with_main_document_and_smart_link_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_test_document_resolved_smart_link_document_list_view()
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(
            response=response, text=self.test_documents[1].label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_main_document_resolved_smart_document_list_with_main_document_and_smart_link_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self.test_documents[0].delete()

        self._clear_events()

        response = self._request_test_document_resolved_smart_link_document_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_resolved_smart_document_list_with_full_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )
        self.grant_access(
            obj=self.test_documents[1], permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_test_document_resolved_smart_link_document_list_view()
        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response=response, text=self.test_documents[1].label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_linked_document_resolved_smart_document_list_with_full_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )
        self.grant_access(
            obj=self.test_documents[1], permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self.test_documents[1].delete()

        self._clear_events()

        response = self._request_test_document_resolved_smart_link_document_list_view()
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(
            response=response, text=self.test_documents[1].label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.contrib import admin

from .models import DuplicateBackendEntry, StoredDuplicateBackend


@admin.register(DuplicateBackendEntry)
class DuplicateBackendEntryAdmin(admin.ModelAdmin):
    list_display = ('document',)


@admin.register(StoredDuplicateBackend)
class StoredDuplicateBackendAdmin(admin.ModelAdmin):
    list_display = (
        'backend_path', 'backend_data'
    )



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('metadata', '0009_auto_20180402_0339'),
    ]

    operations = [
        migrations.AlterField(
            model_name='metadatatype',
            name='default',
            field=models.CharField(
                blank=True,
                help_text="Enter a template to render. Use Django's default "
                "templating language "
                "(https://docs.djangoproject.com/en/1.11/ref/templates/builtins/)",
                max_length=128, null=True, verbose_name='Default'
            ),
        ),
        migrations.AlterField(
            model_name='metadatatype',
            name='lookup',
            field=models.TextField(
                blank=True,
                help_text="Enter a template to render. Must result in a comma "
                "delimited string. Use Django's default templating language "
                "(https://docs.djangoproject.com/en/1.11/ref/templates/builtins/).",
                null=True, verbose_name='Lookup'
            ),
        ),
    ]



from mayan.apps.views.forms import DetailForm

from .models import StoredPermission


class StoredPermissionDetailForm(DetailForm):
    class Meta:
        fields = ('namespace', 'name')
        model = StoredPermission



import os

from django.conf import settings

DEFAULT_SOURCES_BACKEND_ARGUMENTS = {
    'mayan.apps.sources.source_backends.SourceBackendSaneScanner': {
        'scanimage_path': '/usr/bin/scanimage'
    }
}
DEFAULT_SOURCES_CACHE_STORAGE_BACKEND = 'django.core.files.storage.FileSystemStorage'
DEFAULT_SOURCES_CACHE_STORAGE_BACKEND_ARGUMENTS = {
    'location': os.path.join(settings.MEDIA_ROOT, 'source_cache')
}

DEFAULT_SOURCES_LOCK_EXPIRE = 600

STORAGE_NAME_SOURCE_CACHE_FOLDER = 'sources__source_cache'



from django.db import models
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.documents.models import DocumentType
from mayan.apps.events.classes import EventManagerSave
from mayan.apps.events.decorators import method_event
from mayan.apps.templating.classes import Template

from .events import (
    event_web_link_created, event_web_link_edited, event_web_link_navigated
)
from .managers import WebLinkManager


class WebLink(ExtraDataModelMixin, models.Model):
    """
    This model stores the basic fields for a web link. Web links allow
    generating links from documents to external resources.
    """
    label = models.CharField(
        db_index=True, help_text=_('A short text describing the web link.'),
        max_length=96, unique=True, verbose_name=_('Label')
    )
    template = models.TextField(
        help_text=_(
            'Template that will be used to craft the final URL of the '
            'web link.'
        ), verbose_name=_('Template')
    )
    enabled = models.BooleanField(default=True, verbose_name=_('Enabled'))
    document_types = models.ManyToManyField(
        related_name='web_links', to=DocumentType,
        verbose_name=_('Document types')
    )

    class Meta:
        ordering = ('label',)
        verbose_name = _('Web link')
        verbose_name_plural = _('Web links')

    def __str__(self):
        return self.label

    def document_types_add(self, queryset, _event_actor=None):
        for document_type in queryset:
            self.document_types.add(document_type)
            event_web_link_edited.commit(
                action_object=document_type,
                actor=_event_actor or self._event_actor, target=self
            )

    def document_types_remove(self, queryset, _event_actor=None):
        for document_type in queryset:
            self.document_types.remove(document_type)
            event_web_link_edited.commit(
                action_object=document_type,
                actor=_event_actor or self._event_actor, target=self
            )

    def get_absolute_url(self):
        return reverse(
            viewname='web_links:web_link_edit', kwargs={
                'web_link_id': self.pk
            }
        )

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_web_link_created,
            'target': 'self',
        },
        edited={
            'event': event_web_link_edited,
            'target': 'self',
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)


class ResolvedWebLink(WebLink):
    """
    Proxy model to represent an already resolved web link. Used for easier
    colums registration.
    """
    objects = WebLinkManager()

    class Meta:
        proxy = True

    def get_redirect(self, document, user):
        event_web_link_navigated.commit(
            actor=user, action_object=document,
            target=self
        )
        return HttpResponseRedirect(
            redirect_to=self.get_redirect_url_for(document=document)
        )

    def get_redirect_url_for(self, document):
        return Template(template_string=self.template).render(
            context={'document': document}
        )



from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..permissions import permission_document_view

from .mixins.document_mixins import DocumentTestMixin
from .mixins.recently_accessed_document_mixins import (
    RecentlyAccessedDocumentAPIViewTestMixin
)


class RecentlyAccessedDocumentAPIViewTestCase(
    DocumentTestMixin, RecentlyAccessedDocumentAPIViewTestMixin,
    BaseAPITestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_recently_accessed_document_api_list_view_no_activity(self):
        self._clear_events()

        response = self._request_recently_accessed_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_recently_accessed_document_api_list_view_with_activity_no_permission(self):
        self.test_document.add_as_recent_document_for_user(
            user=self._test_case_user
        )

        self._clear_events()

        response = self._request_recently_accessed_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_recently_accessed_document_api_list_view_with_activity_with_access(self):
        self.test_document.add_as_recent_document_for_user(
            user=self._test_case_user
        )

        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_recently_accessed_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['document']['id'],
            self.test_document.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_recently_accessed_document_api_list_view_with_activity_with_access(self):
        self.test_document.add_as_recent_document_for_user(
            user=self._test_case_user
        )

        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_recently_accessed_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import os
import shutil

from mayan.apps.documents.tests.literals import TEST_NON_ASCII_DOCUMENT_PATH
from mayan.apps.storage.utils import mkdtemp
from mayan.apps.testing.tests.base import BaseTestCase

from ..source_backends.staging_folder_backends import StagingFolderFile

from .mocks import MockStagingFolder


class StagingFolderFileTestCase(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.temporary_directory = mkdtemp()
        shutil.copy(
            src=TEST_NON_ASCII_DOCUMENT_PATH, dst=self.temporary_directory
        )
        self.test_filename = os.path.basename(TEST_NON_ASCII_DOCUMENT_PATH)
        self.test_staging_folder = MockStagingFolder()
        self.test_staging_folder.kwargs['folder_path'] = self.temporary_directory
        self.test_staging_folder_files = []

    def tearDown(self):
        for test_staging_folder_file in self.test_staging_folder_files:
            try:
                test_staging_folder_file.delete()
            except FileNotFoundError:
                """Ignore file not found errors."""

        shutil.rmtree(path=self.temporary_directory)
        super().tearDown()

    def test_unicode_staging_folder_file(self):
        self.test_staging_folder_files.append(
            StagingFolderFile(
                staging_folder=self.test_staging_folder,
                filename=self.test_filename
            )
        )

        self.test_staging_folder_files.append(
            StagingFolderFile(
                staging_folder=self.test_staging_folder,
                encoded_filename=self.test_staging_folder_files[0].encoded_filename
            )
        )

        self.assertEqual(
            self.test_staging_folder_files[1].filename, self.test_filename
        )

    def test_staging_folder_file_generate_image_method(self):
        self.test_staging_folder_files.append(
            StagingFolderFile(
                staging_folder=self.test_staging_folder,
                filename=self.test_filename
            )
        )

        self.assertNotEqual(self.test_staging_folder_files[0].generate_image(), '')



import logging

from django.template import Library
from django.utils.encoding import force_text

import mayan

from ..classes import MissingItem
from ..utils import return_attrib

logger = logging.getLogger(name=__name__)
register = Library()


@register.simple_tag
def common_get_missing_items():
    return MissingItem.get_missing()


@register.simple_tag
def common_get_object_verbose_name(obj):
    try:
        return obj._meta.verbose_name
    except AttributeError:
        if isinstance(obj, str):
            return ''
        else:
            return type(obj)


@register.filter
def common_get_type(value):
    return force_text(s=type(value))


@register.filter
def common_object_property(value, arg):
    return return_attrib(value, arg)


@register.simple_tag
def common_project_information(attribute_name):
    return getattr(mayan, attribute_name)



import logging

from django.utils.module_loading import import_string

from .settings import setting_backend, setting_backend_arguments

logger = logging.getLogger(name=__name__)


class MIMETypeBackend:
    @staticmethod
    def get_backend_instance():
        return import_string(dotted_path=setting_backend.value)(
            **setting_backend_arguments.value
        )

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        return self._init(**kwargs)

    def _init(self, **kwargs):
        """Option method for subclasses to overload."""

    def get_mime_type(self, file_object, mime_type_only=False):
        return self._get_mime_type(
            file_object=file_object, mime_type_only=mime_type_only
        )



from django.conf.urls import url

from .api_views.index_instance_api_views import (
    APIDocumentIndexInstanceNodeListView, APIIndexInstanceDetailView,
    APIIndexInstanceListView, APIIndexInstanceNodeChildrenNodeListView,
    APIIndexInstanceNodeDetailView, APIIndexInstanceNodeDocumentListView,
    APIIndexInstanceNodeListView
)
from .api_views.index_template_api_views import (
    APIIndexTemplateDetailView, APIIndexTemplateDocumentTypeAddView,
    APIIndexTemplateDocumentTypeListView,
    APIIndexTemplateDocumentTypeRemoveView, APIIndexTemplateListView,
    APIIndexTemplateNodeListView, APIIndexTemplateNodeDetailView,
    APIIndexTemplateRebuildView, APIIndexTemplateResetView
)

from .views.index_instance_views import (
    DocumentIndexInstanceNodeListView, IndexInstanceNodeView,
    IndexInstanceListView
)
from .views.index_template_views import (
    DocumentTypeIndexTemplateListView, IndexTemplateAllRebuildView,
    IndexTemplateCreateView, IndexTemplateDeleteView,
    IndexTemplateDocumentTypesView, IndexTemplateEditView,
    IndexTemplateListView, IndexTemplateNodeListView,
    IndexTemplateNodeCreateView, IndexTemplateNodeDeleteView,
    IndexTemplateNodeEditView, IndexTemplateRebuildView,
    IndexTemplateResetView
)

urlpatterns_templates = [
    url(
        regex=r'^document_types/(?P<document_type_id>\d+)/index_templates/$',
        name='document_type_index_templates',
        view=DocumentTypeIndexTemplateListView.as_view()
    ),
    url(
        regex=r'^templates/$', name='index_template_list',
        view=IndexTemplateListView.as_view()
    ),
    url(
        regex=r'^templates/create/$', name='index_template_create',
        view=IndexTemplateCreateView.as_view()
    ),
    url(
        regex=r'^templates/(?P<index_template_id>\d+)/delete/$',
        name='index_template_delete', view=IndexTemplateDeleteView.as_view()
    ),
    url(
        regex=r'^templates/(?P<index_template_id>\d+)/document_types/$',
        name='index_template_document_types',
        view=IndexTemplateDocumentTypesView.as_view()
    ),
    url(
        regex=r'^templates/(?P<index_template_id>\d+)/edit/$',
        name='index_template_edit', view=IndexTemplateEditView.as_view()
    ),
    url(
        regex=r'^templates/(?P<index_template_id>\d+)/nodes/$',
        name='index_template_view',
        view=IndexTemplateNodeListView.as_view()
    ),
    url(
        regex=r'^templates/(?P<index_template_id>\d+)/rebuild/$',
        name='index_template_rebuild', view=IndexTemplateRebuildView.as_view()
    ),
    url(
        regex=r'^templates/nodes/(?P<index_template_node_id>\d+)/children/create/$',
        name='template_node_create', view=IndexTemplateNodeCreateView.as_view()
    ),
    url(
        regex=r'^templates/nodes/(?P<index_template_node_id>\d+)/delete/$',
        name='template_node_delete', view=IndexTemplateNodeDeleteView.as_view()
    ),
    url(
        regex=r'^templates/nodes/(?P<index_template_node_id>\d+)/edit/$',
        name='template_node_edit', view=IndexTemplateNodeEditView.as_view()
    )
]

urlpatterns_instances = [
    url(
        regex=r'^instances/$', name='index_list', view=IndexInstanceListView.as_view()
    ),
    url(
        regex=r'^instances/nodes/(?P<index_instance_node_id>\d+)/$',
        name='index_instance_node_view', view=IndexInstanceNodeView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>\d+)/instances/$',
        name='document_index_list', view=DocumentIndexInstanceNodeListView.as_view()
    )
]

urlpatterns_tools = [
    url(
        regex=r'^instances/rebuild/$', name='rebuild_index_instances',
        view=IndexTemplateAllRebuildView.as_view()
    ),
    url(
        regex=r'^instances/reset/$', name='index_instances_reset',
        view=IndexTemplateResetView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_templates)
urlpatterns.extend(urlpatterns_instances)
urlpatterns.extend(urlpatterns_tools)

api_urls_document_indexes = [
    url(
        regex=r'^documents/(?P<document_id>[0-9]+)/indexes/$',
        name='document-index-list',
        view=APIDocumentIndexInstanceNodeListView.as_view()
    ),
]

api_urls_index_instances = [
    url(
        regex=r'^index_instances/$', name='indexinstance-list',
        view=APIIndexInstanceListView.as_view()
    ),
    url(
        regex=r'^index_instances/(?P<index_instance_id>[0-9]+)/$',
        name='indexinstance-detail', view=APIIndexInstanceDetailView.as_view()
    ),
    url(
        regex=r'^index_instances/(?P<index_instance_id>[0-9]+)/nodes/$',
        name='indexinstancenode-list', view=APIIndexInstanceNodeListView.as_view()
    ),
    url(
        regex=r'^index_instances/(?P<index_instance_id>[0-9]+)/nodes/(?P<index_instance_node_id>[0-9]+)/$',
        name='indexinstancenode-detail',
        view=APIIndexInstanceNodeDetailView.as_view()
    ),
    url(
        regex=r'^index_instances/(?P<index_instance_id>[0-9]+)/nodes/(?P<index_instance_node_id>[0-9]+)/nodes/$',
        name='indexinstancenode-children-list',
        view=APIIndexInstanceNodeChildrenNodeListView.as_view()
    ),
    url(
        regex=r'^index_instances/(?P<index_instance_id>[0-9]+)/nodes/(?P<index_instance_node_id>[0-9]+)/documents/$',
        name='indexinstancenode-document-list',
        view=APIIndexInstanceNodeDocumentListView.as_view()
    ),
]

api_urls_index_templates = [
    url(
        regex=r'^index_templates/$', name='indextemplate-list',
        view=APIIndexTemplateListView.as_view()
    ),
    url(
        regex=r'^index_templates/(?P<index_template_id>[0-9]+)/$',
        name='indextemplate-detail',
        view=APIIndexTemplateDetailView.as_view()
    ),
    url(
        regex=r'^index_templates/(?P<index_template_id>[0-9]+)/document_types/$',
        name='indextemplate-documenttype-list',
        view=APIIndexTemplateDocumentTypeListView.as_view()
    ),
    url(
        regex=r'^index_templates/(?P<index_template_id>[0-9]+)/document_types/add/$',
        name='indextemplate-documenttype-add',
        view=APIIndexTemplateDocumentTypeAddView.as_view()
    ),
    url(
        regex=r'^index_templates/(?P<index_template_id>[0-9]+)/document_types/remove/$',
        name='indextemplate-documenttype-remove',
        view=APIIndexTemplateDocumentTypeRemoveView.as_view()
    ),
    url(
        regex=r'^index_templates/(?P<index_template_id>[0-9]+)/rebuild/$',
        name='indextemplate-rebuild',
        view=APIIndexTemplateRebuildView.as_view()
    ),
    url(
        regex=r'^index_templates/(?P<index_template_id>[0-9]+)/reset/$',
        name='indextemplate-reset',
        view=APIIndexTemplateResetView.as_view()
    ),
    url(
        regex=r'^index_templates/(?P<index_template_id>[0-9]+)/nodes/$',
        name='indextemplatenode-list',
        view=APIIndexTemplateNodeListView.as_view()
    ),
    url(
        regex=r'^index_templates/(?P<index_template_id>[0-9]+)/nodes/(?P<index_template_node_id>[0-9]+)/$',
        name='indextemplatenode-detail', view=APIIndexTemplateNodeDetailView.as_view()
    ),
]

api_urls = []
api_urls.extend(api_urls_document_indexes)
api_urls.extend(api_urls_index_instances)
api_urls.extend(api_urls_index_templates)




class FileMetadataError(Exception):
    """Base file metadata driver exception."""


class FileMetadataDriverError(FileMetadataError):
    """Exception raised when a driver encounters an unexpected error."""
